﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Core.Networks;
using HeuristicLab.Data;
using HeuristicLab.Encodings.RealVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.FacilityLocation;
using HeuristicLab.Problems.VehicleRouting;
using HeuristicLab.Problems.VehicleRouting.Encodings.General;
using HeuristicLab.Problems.VehicleRouting.ProblemInstances;

namespace HeuristicLab.Networks.IntegratedOptimization.LocationRouting {
  [Item("LrpOrchestratorNode", "An abstract base class for orchestrators used in LRP optimization networks.")]
  [StorableClass]
  public abstract class LrpOrchestratorNode : OrchestratorNode {
    #region Constants
    protected const string InstanceParameterName = "Instance";
    protected const string FlpParameterName = "FLP";
    protected const string VrpParameterName = "VRP";
    protected const string MetaSolverName = "MetaSolver";
    protected const string FlpSolverName = "FlpSolver";
    protected const string VrpSolverName = "VrpSolver";
    protected const string OrchestrationPortNameSuffix = "OrchestrationPort";
    protected const string EvaluationPortNameSuffix = "EvaluationPort";
    #endregion

    protected CancellationTokenSource cts;

    [Storable]
    protected int nrOfDepots, nrOfCustomers;
    [Storable]
    protected double[,] depotCoordinates = new double[0, 0], customerCoordinates = new double[0, 0];
    [Storable]
    protected LrpUtils.DistanceType distanceType;
    [Storable]
    protected double[] depotCapacities = new double[0], customerDemands = new double[0], depotCosts = new double[0];
    [Storable]
    protected double vehicleCapacity, vehicleCost;

    #region Parameters
    public IFixedValueParameter<TextFileValue> InstanceParameter {
      get { return (IFixedValueParameter<TextFileValue>)Parameters[InstanceParameterName]; }
    }

    public IValueParameter<FacilityLocationProblem> FlpParameter {
      get { return (IValueParameter<FacilityLocationProblem>)Parameters[FlpParameterName]; }
    }

    public IValueParameter<VehicleRoutingProblem> VrpParameter {
      get { return (IValueParameter<VehicleRoutingProblem>)Parameters[VrpParameterName]; }
    }
    #endregion

    #region Ports
    public IMessagePort MetaSolverOrchestrationPort {
      get { return (IMessagePort)Ports[MetaSolverName + OrchestrationPortNameSuffix]; }
      protected set {
        string portName = MetaSolverName + OrchestrationPortNameSuffix;
        if (Ports.ContainsKey(portName))
          throw new InvalidOperationException(portName + " already set.");
        Ports.Add(value);
        value.ConnectedPortChanged += MetaSolverOrchestrationPort_ConnectedPortChanged;
      }
    }

    public IMessagePort MetaSolverEvaluationPort {
      get { return (IMessagePort)Ports[MetaSolverName + EvaluationPortNameSuffix]; }
      protected set {
        string portName = MetaSolverName + EvaluationPortNameSuffix;
        if (Ports.ContainsKey(portName))
          throw new InvalidOperationException(portName + " already set.");
        Ports.Add(value);
      }
    }

    public IMessagePort FlpSolverOrchestrationPort {
      get { return (IMessagePort)Ports[FlpSolverName + OrchestrationPortNameSuffix]; }
      protected set {
        string portName = FlpSolverName + OrchestrationPortNameSuffix;
        if (Ports.ContainsKey(portName))
          throw new InvalidOperationException(portName + " already set.");
        Ports.Add(value);
        value.ConnectedPortChanged += FlpSolverOrchestrationPort_ConnectedPortChanged;
      }
    }

    public IMessagePort FlpSolverEvaluationPort {
      get { return (IMessagePort)Ports[FlpSolverName + EvaluationPortNameSuffix]; }
      protected set {
        string portName = FlpSolverName + EvaluationPortNameSuffix;
        if (Ports.ContainsKey(portName))
          throw new InvalidOperationException(portName + " already set.");
        Ports.Add(value);
      }
    }

    public IMessagePort VrpSolverOrchestrationPort {
      get { return (IMessagePort)Ports[VrpSolverName + OrchestrationPortNameSuffix]; }
      protected set {
        string portName = VrpSolverName + OrchestrationPortNameSuffix;
        if (Ports.ContainsKey(portName))
          throw new InvalidOperationException(portName + " already set.");
        Ports.Add(value);
        value.ConnectedPortChanged += VrpSolverOrchestrationPort_ConnectedPortChanged;
      }
    }

    public IMessagePort VrpSolverEvaluationPort {
      get { return (IMessagePort)Ports[VrpSolverName + EvaluationPortNameSuffix]; }
      protected set {
        string portName = VrpSolverName + EvaluationPortNameSuffix;
        if (Ports.ContainsKey(portName))
          throw new InvalidOperationException(portName + " already set.");
        Ports.Add(value);
      }
    }
    #endregion

    [StorableConstructor]
    protected LrpOrchestratorNode(bool deserializing) : base(deserializing) { }
    protected LrpOrchestratorNode(LrpOrchestratorNode original, Cloner cloner) : base(original, cloner) {
      nrOfDepots = original.nrOfDepots;
      nrOfCustomers = original.nrOfCustomers;
      depotCoordinates = (double[,])original.depotCoordinates.Clone();
      customerCoordinates = (double[,])original.customerCoordinates.Clone();
      distanceType = original.distanceType;
      depotCapacities = (double[])original.depotCapacities.Clone();
      customerDemands = (double[])original.customerDemands.Clone();
      depotCosts = (double[])original.depotCosts.Clone();
      vehicleCapacity = original.vehicleCapacity;
      vehicleCost = original.vehicleCost;

      RegisterEvents();
    }
    protected LrpOrchestratorNode() : this("LrpOrchestratorNode") { }
    protected LrpOrchestratorNode(string name) : base(name) {
      Parameters.Add(new FixedValueParameter<TextFileValue>(InstanceParameterName));
      Parameters.Add(new ValueParameter<FacilityLocationProblem>(FlpParameterName, new FacilityLocationProblem()));
      Parameters.Add(new ValueParameter<VehicleRoutingProblem>(VrpParameterName, new VehicleRoutingProblem() { ProblemInstance = new CVRPProblemInstance() }));

      InstanceParameter.Value.ToStringChanged += InstanceParameter_Value_ToStringChanged;
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      RegisterEvents();
    }

    private void RegisterEvents() {
      InstanceParameter.Value.ToStringChanged += InstanceParameter_Value_ToStringChanged;
      MetaSolverOrchestrationPort.ConnectedPortChanged += MetaSolverOrchestrationPort_ConnectedPortChanged;
      FlpSolverOrchestrationPort.ConnectedPortChanged += FlpSolverOrchestrationPort_ConnectedPortChanged;
      VrpSolverOrchestrationPort.ConnectedPortChanged += VrpSolverOrchestrationPort_ConnectedPortChanged;
    }

    protected virtual void InstanceParameter_Value_ToStringChanged(object sender, EventArgs e) {
      string filePath = InstanceParameter.Value.Value;
      LrpUtils.Import(filePath, out nrOfDepots, out nrOfCustomers,
                                out depotCoordinates, out customerCoordinates,
                                out distanceType,
                                out depotCapacities, out customerDemands, out depotCosts,
                                out vehicleCapacity, out vehicleCost);

      var flp = FlpParameter.Value;
      flp.CustomerDemandsParameter.Value = new DoubleArray(customerDemands);
      flp.DeliveryCostsParameter.Value = new DoubleMatrix(LrpUtils.GetFlpDeliveryCosts(depotCoordinates, customerCoordinates, distanceType));
      flp.DepotCapacitiesParameter.Value = new DoubleArray(depotCapacities);
      flp.Encoding.Length = nrOfCustomers;
      flp.OpeningCostsParameter.Value = new DoubleArray(depotCosts);

      var vrpInstance = (CVRPProblemInstance)VrpParameter.Value.ProblemInstance;
      vrpInstance.Capacity.Value = vehicleCapacity;
      vrpInstance.FleetUsageFactor.Value = vehicleCost;
      vrpInstance.OverloadPenalty.Value = vehicleCost * 1000.0;

      var crossover = VrpParameter.Value.OperatorsParameter.Value.OfType<MultiVRPSolutionCrossover>().Single(x => x.Name == "MultiVRPSolutionCrossover");
      foreach (var c in crossover.Operators)
        crossover.Operators.SetItemCheckedState(c, c.Name.StartsWith("Potvin"));
      var mutator = VrpParameter.Value.OperatorsParameter.Value.OfType<MultiVRPSolutionManipulator>().Single(x => x.Name == "MultiVRPSolutionManipulator");
      foreach (var m in mutator.Operators)
        mutator.Operators.SetItemCheckedState(m, Regex.IsMatch(m.Name, @"Potvin(One|Two).*"));

      Prepare();
    }

    protected virtual void MetaSolverOrchestrationPort_ConnectedPortChanged(object sender, EventArgs e) {
      if (MetaSolverOrchestrationPort.ConnectedPort == null) return;

      var node = MetaSolverOrchestrationPort.ConnectedPort.Parent as OrchestratedAlgorithmNode;
      if (node == null) return;

      var hook = new HookOperator { Name = "Meta Eval Hook" };
      hook.Parameters.Add(new LookupParameter<RealVector>("RealVector") { Hidden = true });
      hook.Parameters.Add(new LookupParameter<DoubleValue>("Quality") { Hidden = true });
      node.EvalHook = hook;

      node.OrchestrationPort.CloneParametersFromPort(MetaSolverOrchestrationPort);
      node.EvaluationPort.CloneParametersFromPort(MetaSolverEvaluationPort);
      node.EvaluationPort.ConnectedPort = MetaSolverEvaluationPort;
    }

    protected virtual void FlpSolverOrchestrationPort_ConnectedPortChanged(object sender, EventArgs e) {
      if (FlpSolverOrchestrationPort.ConnectedPort == null) return;

      var node = FlpSolverOrchestrationPort.ConnectedPort.Parent as OrchestratedAlgorithmNode;
      if (node == null) return;

      node.OrchestrationPort.CloneParametersFromPort(FlpSolverOrchestrationPort);
    }

    protected virtual void VrpSolverOrchestrationPort_ConnectedPortChanged(object sender, EventArgs e) {
      if (VrpSolverOrchestrationPort.ConnectedPort == null) return;

      var node = VrpSolverOrchestrationPort.ConnectedPort.Parent as OrchestratedAlgorithmNode;
      if (node == null) return;

      node.OrchestrationPort.CloneParametersFromPort(VrpSolverOrchestrationPort);
    }

    public override void Prepare(bool clearRuns = false) {
      Results.Clear();
    }

    public override void Start() {
      cts = new CancellationTokenSource();

      var metaMsg = MetaSolverOrchestrationPort.PrepareMessage();
      metaMsg["OrchestrationMessage"] = new EnumValue<OrchestrationMessage>(OrchestrationMessage.Start);
      MetaSolverOrchestrationPort.SendMessage(metaMsg);
    }

    public override void Pause() {
      cts.Cancel();

      var metaMsg = MetaSolverOrchestrationPort.PrepareMessage();
      metaMsg["OrchestrationMessage"] = new EnumValue<OrchestrationMessage>(OrchestrationMessage.Pause);
      MetaSolverOrchestrationPort.SendMessage(metaMsg);

      var flpMsg = FlpSolverOrchestrationPort.PrepareMessage();
      flpMsg["OrchestrationMessage"] = new EnumValue<OrchestrationMessage>(OrchestrationMessage.Stop);
      FlpSolverOrchestrationPort.SendMessage(flpMsg);

      var vrpMsg = VrpSolverOrchestrationPort.PrepareMessage();
      vrpMsg["OrchestrationMessage"] = new EnumValue<OrchestrationMessage>(OrchestrationMessage.Stop);
      VrpSolverOrchestrationPort.SendMessage(vrpMsg);
    }

    public override void Stop() {
      cts.Cancel();

      var metaMsg = MetaSolverOrchestrationPort.PrepareMessage();
      metaMsg["OrchestrationMessage"] = new EnumValue<OrchestrationMessage>(OrchestrationMessage.Stop);
      MetaSolverOrchestrationPort.SendMessage(metaMsg);

      var flpMsg = FlpSolverOrchestrationPort.PrepareMessage();
      flpMsg["OrchestrationMessage"] = new EnumValue<OrchestrationMessage>(OrchestrationMessage.Stop);
      FlpSolverOrchestrationPort.SendMessage(flpMsg);

      var vrpMsg = VrpSolverOrchestrationPort.PrepareMessage();
      vrpMsg["OrchestrationMessage"] = new EnumValue<OrchestrationMessage>(OrchestrationMessage.Stop);
      VrpSolverOrchestrationPort.SendMessage(vrpMsg);
    }

    protected override void ProcessMessage(IMessage message, IMessagePort port, CancellationToken token) {
      var messageActions = new Dictionary<IMessagePort, Action<IMessage>>();
      messageActions.Add(MetaSolverOrchestrationPort, MetaSolverOrchestrationPortMessage);
      messageActions.Add(MetaSolverEvaluationPort, MetaSolverEvaluationPortMessage);
      messageActions.Add(FlpSolverOrchestrationPort, FlpSolverOrchestrationPortMessage);
      messageActions.Add(VrpSolverOrchestrationPort, VrpSolverOrchestrationPortMessage);

      messageActions[port](message);

      base.ProcessMessage(message, port, token);
    }

    #region MetaSolver Message Handling
    protected virtual void MetaSolverOrchestrationPortMessage(IMessage message) { }

    protected virtual void MetaSolverEvaluationPortMessage(IMessage message) { }
    #endregion

    #region FlpSolver Message Handling
    protected virtual void FlpSolverOrchestrationPortMessage(IMessage message) { }

    protected virtual void FlpSolverEvaluationPortMessage(IMessage message) { }
    #endregion

    #region VrpSolver Message Handling
    protected virtual void VrpSolverOrchestrationPortMessage(IMessage message) { }

    private void VrpSolverEvaluationPortMessage(IMessage message) { }
    #endregion
  }
}
